////////////////////////////////////////////////////////////////////////////////////
/// @file CModel.cpp
/// @author DGI
/// @date 2005-06-15
/// @version 1.0
///
/// @addtogroup chargeur3ds Chargeur 3DS
/// @{
////////////////////////////////////////////////////////////////////////////////////

#ifdef WIN32
#include <windows.h>
#endif

#include <GL/gl.h>

#include "CModel.h"
#include "CObjet.h"
#include "CMateriel.h"
#include "CFace.h"

#include "Utilitaire.h"

namespace chargeur3ds {


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn operator< ( CEcritureFichierBinaire& out, const CModel& mod)
   ///
   /// Surcharge de l'oprateur <
   ///
   /// @param[out] out : l'identificateur du fichier
   /// @param[in] mod  : le modle
   ///
   /// @return out : l'identificateur du fichier
   ///
   ////////////////////////////////////////////////////////////////////////
   CEcritureFichierBinaire& operator< ( CEcritureFichierBinaire& out, const CModel& mod)
   {
       out<int ( mod.vecteurObjets_.size());
       for ( unsigned int i = 0; i < mod.vecteurObjets_.size(); i++)  {
           out<mod.vecteurObjets_[i];
       }
       out<int ( mod.vecteurMateriaux_.size());
       for ( unsigned int i = 0; i < mod.vecteurMateriaux_.size(); i++) {
           out<mod.vecteurMateriaux_[i];
       }
       return out;
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn operator> ( CLectureFichierBinaire& in, CModel& mod)
   ///
   /// Description : Surcharge de l'oprateur >
   ///
   /// @param[in] in   : l'identificateur du fichier
   /// @param[out] mod : le modle
   ///
   /// @return in : l'identificateur du fichier 
   ///
   ////////////////////////////////////////////////////////////////////////
   CLectureFichierBinaire& operator> ( CLectureFichierBinaire& in, CModel& mod)
   {
       // Lire les objects
       unsigned int size;
       in>size;
       mod.vecteurObjets_.resize ( size);
       for ( unsigned int i = 0; i < mod.vecteurObjets_.size(); i++)     {
           in>mod.vecteurObjets_[i];
       }

       // Lire les matriaux
       in>size;
       mod.vecteurMateriaux_.resize ( size);
       for ( unsigned int i = 0; i < mod.vecteurMateriaux_.size(); i++) {
           in>mod.vecteurMateriaux_[i];
       }

       return in;
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::CModel()
   ///
   /// Constructeur par dfaut
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   CModel::CModel()
   {
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::~CModel()
   ///
   /// Destructeur par dfaut
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   CModel::~CModel()
   {
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::exprimerParRapportAuPivot()
   ///
   /// Effectue une translation de vecteur-pivot
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::exprimerParRapportAuPivot()
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           vecteurObjets_[i].Translate ( vecteurObjets_[i].positionPivot* ( -1));
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::importer ( const char* fname)
   ///
   /// Charge un modle  partir d'un fichier texte
   ///
   /// @param[in] fname : nom du fichier
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::importer ( const char* fname)
   {
       vecteurObjets_.clear();
       vecteurMateriaux_.clear();

       ifstream in ( fname);
       if  ( in.fail()) {
          utilitaire::afficherErreur(
             string("Erreur a l'ouverture du fichier \"") +
             string(fname) + string("\" en lecture.")
             );
          return;
       }

       string r;
       in>>r;
       int n_mat;
       in>>n_mat;

       for ( int i = 0; i < n_mat; i++)  {
           vecteurMateriaux_.push_back ( CMateriel());
           vecteurMateriaux_.back().ImportMaterial ( in);
       }

       int n_obj;
       in>>r;
       in>>n_obj;


       for ( int i = 0;  i < n_obj; i++) {
           vecteurObjets_.push_back ( CObjet());
           vecteurObjets_.back().importer ( in);
           //    vecteurObjets_.back().ptrMateriel=&vecteurMateriaux_[vecteurObjets_.back().mat_id];
       }


       in.close();
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::Load ( string nom_modele, const char* fname, const char* relatif = "." )
   ///
   /// Charge le modle,  partir du fichier .bin s'il existe, sinon  partir
   /// du fichier texte, et cre le .bin.
   ///
   /// @param nom_modele : nom du modle
   /// @param fname      : nom du fichier 
   /// @param relatif    : rpertoire relatif du modle par rapport au rpertoire courant
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::Load ( string nom_modele, const char* fname, const char* relatif)
   {
       nom_ = nom_modele;

       relatif_ = relatif;

       string sname( relatif_ );
       
       sname += string("/") + fname;

       string stxt = sname + ".txt";
       string sbin = sname + ".bin";


       CLectureFichierBinaire in ( sbin.c_str());
       if  ( !in.fail())  {
           in > (*this);
           in.close();
       }
       else   {

           importer ( stxt.c_str());
           Write ( sbin.c_str());
       }

       ecrirePointeurMateriel();

   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::Write ( const char* fname) const
   ///
   /// crit le modle sous forme de fichier binaire
   ///
   /// @param[in] fname : le nom du fichier
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::Write ( const char* fname) const
   {
       CEcritureFichierBinaire out ( fname);
       if (out.fail()) {
         utilitaire::afficherErreur(
            string("Erreur a l'ouverture du fichier \"") +
            string(fname) + string("\" en ecriture.")
            );
         return;
       }
       out< (*this);
       out.close();
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::ecrirePointeurMateriel()
   ///
   /// Met  jour le pointeur matriel pour tous les objets
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::ecrirePointeurMateriel()
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++)  {
         if (vecteurObjets_[i].mat_id != -1)
         {
            vecteurObjets_[i].ptrMateriel = &vecteurMateriaux_[vecteurObjets_[i].mat_id];
         }
         else
         {
            vecteurObjets_[i].ptrMateriel = CMateriel::GetSansMateriel();
         }
      }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::mettreAEchelle ( float f)
   ///
   /// Mise  l'chelle du modle
   ///
   /// @param[in] f : Facteur de mise  l'chelle du modle
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::mettreAEchelle ( float f)
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           vecteurObjets_[i].mettreAEchelle ( f);
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::tourner( float angle , const Vecteur3& axe )
   ///
   /// Rotation du modle
   ///
   /// @param[in] angle : Angle de rotation, en radians
   /// @param[in] axe   : Axe de rotation (unitaire)
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::tourner( float angle , const Vecteur3& axe )
   {
      for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
         vecteurObjets_[i].tourner(angle, axe);
      }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::dessinerPoint() const
   ///
   /// Dessine la gomtrie sous forme de points
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::dessinerPoint() const
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           const CObjet& obj = vecteurObjets_[i];

           glBegin ( GL_POINTS);
           for ( unsigned int j = 0; j < obj.vvert.size(); j++) {
               Vecteur3 pos = obj.vvert[j].pos;
               glVertex3d ( pos[0], pos[1], pos[2]);
           }
           glEnd();
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::dessiner() const
   ///
   /// Dessine la gomtrie (en activant le bon matriel pour chaque objet)
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::dessiner() const
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
          const CObjet& obj = vecteurObjets_[i];

          glPushAttrib(GL_CURRENT_BIT);

          if (obj.mat_id != -1)
          {
             if ( vecteurMateriaux_[obj.mat_id].Reflective() ) continue;
             vecteurMateriaux_[obj.mat_id].Activate();
          }
          else
          {
             CMateriel::GetSansMateriel()->Activate();
          }

          glBegin ( GL_TRIANGLES);
          for ( unsigned int j = 0; j < obj.vface.size(); j++) {
             Vecteur3 fnormal = obj.vface[j].normal;
             glNormal3d ( fnormal[0], fnormal[1], fnormal[2]);
             for ( int k = 0; k < 3; k++) {
                int vid  = obj.vface[j].id[k];
                int tcid = obj.vface[j].tid[k];
                
                Vecteur3 vnormal = obj.vface[j].vnormal[k];
                
                //                    if  ( dot ( vnormal,fnormal)>.9f)
                {
                   glNormal3dv ( &vnormal[0]);
                }

                const Vecteur2f& tcoord = obj.vtc[tcid].tc;
                glTexCoord2fv ( &tcoord[0]);
                const Vecteur3&  pos    = obj.vvert[vid].pos;
                glVertex3dv ( &pos[0]);
             }
          }
          glEnd();
          
          glPopAttrib();
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::chargerMateriaux()
   ///
   /// Charge les matriaux en mmoire opengl
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::chargerMateriaux()
   {
       for ( unsigned int i = 0; i < vecteurMateriaux_.size(); i++) {
           vecteurMateriaux_[i].Load(relatif_.c_str());
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::effacerMateriaux()
   ///
   /// Efface les matriaux
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::effacerMateriaux()
   {
       for ( unsigned int i = 0; i < vecteurMateriaux_.size(); i++)  {
           vecteurMateriaux_[i].Remove();
       }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::dessinerObjet ( int id) const
   ///
   /// Dessine un objet en particulier
   ///
   /// @param[in] id : index de l'objet a dessiner
   ///
   /// @return Aucune
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::dessinerObjet ( int id) const
   {
      const CObjet& obj = vecteurObjets_[id];

      if (obj.mat_id != -1)
      {
         vecteurMateriaux_[obj.mat_id].Activate();
      }
      else
      {
         CMateriel::GetSansMateriel()->Activate();
      }
      
      glBegin ( GL_TRIANGLES);
      for ( unsigned int j = 0; j < obj.vface.size(); j++)   {
         Vecteur3 fnormal = obj.vface[j].normal;
         glNormal3d ( fnormal[0], fnormal[1], fnormal[2]);
         for ( int k = 0; k < 3; k++)   {
            int vid  = obj.vface[j].id[k];
            int tcid = obj.vface[j].tid[k];
            
            Vecteur3 vnormal = obj.vface[j].vnormal[k];

            //                    if  ( dot ( vnormal,fnormal)>.9f)
            {
               glNormal3dv ( &vnormal[0]);
            }
            
            const Vecteur2f& tcoord = obj.vtc[tcid].tc;
            glTexCoord2fv ( &tcoord[0]);
            const Vecteur3&  pos    = obj.vvert[vid].pos;
            glVertex3dv ( &pos[0]);
         }
      }
      glEnd();
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::GetReflectiveId() const
   ///
   /// Retourne la liste des index des objets rflchissants
   ///
   /// @return Retourne la liste des index des objets rflchissants
   ///
   ////////////////////////////////////////////////////////////////////////
   vector<int> CModel::GetReflectiveId() const
   {
       vector<int> r;
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           const CObjet& obj = vecteurObjets_[i];
           if  ( obj.mat_id != -1 && vecteurMateriaux_[obj.mat_id].Reflective()) r.push_back ( i);
       }
       return r;
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::chercher ( const string& nom_objet)
   ///
   /// Cherche l'objet nomm nom_objet
   ///
   /// @param nom_objet : string& nom_objet
   ///
   /// @return l'objet recherche
   ///
   ////////////////////////////////////////////////////////////////////////
   CObjet* CModel::chercher ( const string& nom_objet)
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           //cerr<<vecteurObjets_[i].name<<endl;
           if ( nom_objet == vecteurObjets_[i].name )  {
               return &vecteurObjets_[i];
           }
       }
       return 0;
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn CModel::chercher ( const string& nom_objet) const
   ///
   /// Cherche l'objet nomm nom_objet
   ///
   /// @param nom_objet : string& nom_objet
   ///
   /// @return l'objet recherche
   ///
   ////////////////////////////////////////////////////////////////////////
   const CObjet* CModel::chercher ( const string& nom_objet) const
   {
       for ( unsigned int i = 0; i < vecteurObjets_.size(); i++) {
           //cerr<<vecteurObjets_[i].name<<endl;
           if ( nom_objet == vecteurObjets_[i].name )  {
               return &vecteurObjets_[i];
           }
       }
       return 0;
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   ///    @fn double CObjet::calculerSphereEnglobante() const
   ///
   ///    Description : Calcule et retourne une sphre englobant le modle
   ///
   ///    @return Le rayon de la sphre englobante.
   ///
   ////////////////////////////////////////////////////////////////////////
   double CModel::calculerSphereEnglobante() const
   {
      if (vecteurObjets_.size() == 0) {
         // Aucun sous-objet
         return 0;
      }
      else {
         double rayon = vecteurObjets_[0].calculerSphereEnglobante();

         // Si un des sous-objets n'a pas de vertex, il retournera 0 ce qui
         // pourrait tre problmatique.
         for (unsigned int i = 1; i < vecteurObjets_.size(); i++) {
            double r = vecteurObjets_[i].calculerSphereEnglobante();

            if (rayon < r)
               rayon = r;
         }

         return rayon;
      }
   }

   ////////////////////////////////////////////////////////////////////////
   ///
   /// @fn void CModel::calculerCylindreEnglobant(double& rayon, double& bas, double& haut) const
   ///
   /// Description : Calcule et retourne un cylindre englobant le modle
   ///
   /// @param[out] rayon : Le rayon du cylindre.
   /// @param[out] bas   : La hauteur du bas du cylindre.
   /// @param[out] haut  : La hauteur du haut du cylindre.
   ///
   /// @return Aucune.
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::calculerCylindreEnglobant(double& rayon, double& bas, double& haut) const
   {
      if (vecteurObjets_.size() == 0) {
         // Aucun sous-objet
         rayon = bas = haut = 0;
      }
      else {
         vecteurObjets_[0].calculerCylindreEnglobant(
            rayon, bas, haut
            );

         // Si un des sous-objets n'a pas de vertex, il retournera 0 ce qui
         // pourrait tre problmatique.
         for (unsigned int i = 1; i < vecteurObjets_.size(); i++) {
            double r, b, h;
            vecteurObjets_[i].calculerCylindreEnglobant(r, b, h);

            if (rayon < r)
               rayon = r;
            if (bas > b)
               bas = b;
            if (haut < h)
               haut = h;
         }
      }
   }


   ////////////////////////////////////////////////////////////////////////
   ///
   ///    @fn CModel::calculerBoiteEnglobante(Vecteur3& coinMin, Vecteur3& coinMax) const
   ///
   ///    Description : Calcule et retourne une bote englobante de l'objet.
   ///
   /// @param[out] coinMin : Le coin de coordonnes minimales de la bote englobante.
   /// @param[out] coinMax : Le coin de coordonnes maximales de la bote englobante.
   ///
   ///    @return Aucune.
   ///
   ////////////////////////////////////////////////////////////////////////
   void CModel::calculerBoiteEnglobante(Vecteur3& coinMin, Vecteur3& coinMax) const
   {
      if (vecteurObjets_.size() == 0) {
         // Aucun sous-objet
         coinMin = coinMax = Vecteur3(0);
      }
      else {
         vecteurObjets_[0].calculerBoiteEnglobante(
            coinMin, coinMax
            );

         // Si un des sous-objets n'a pas de vertex, il retournera 0 ce qui
         // pourrait tre problmatique.
         for (unsigned int i = 1; i < vecteurObjets_.size(); i++) {
            Vecteur3 v1, v2;
            vecteurObjets_[i].calculerBoiteEnglobante(v1, v2);

            // Comparaison avec le coin minimal
            if (coinMin[0] > v1[0])
               coinMin[0] = v1[0];
            if (coinMin[1] > v1[1])
               coinMin[1] = v1[1];
            if (coinMin[2] > v1[2])
               coinMin[2] = v1[2];

            // Comparaison avec le coin maximal
            if (coinMax[0] < v2[0])
               coinMax[0] = v2[0];
            if (coinMax[1] < v2[1])
               coinMax[1] = v2[1];
            if (coinMax[2] < v2[2])
               coinMax[2] = v2[2];
         }
      }
   }


} // Fin de l'espace de nom chargeur3ds


////////////////////////////////////////////////////////////////////////////////////
/// @}
////////////////////////////////////////////////////////////////////////////////////
